from types import StringType, ListType
from oldowan.polymorphism import Polymorphism

def sites2str(sites):
    """Transform a list of Polymorphisms to a string.

    """
    processing = []
    for x in sites:
        # unnested (unambiguous) sites get directly pushed to the string
        if isinstance(x,Polymorphism):
            processing.append(str(x))
        # ambiguous sites are nested in lists
        elif type(x) == ListType:
            current = []
            for y in x:
                if type(y) == ListType:
                    if len(y) == 1:
                        current.append(str(y[0]))
                    elif len(y) > 1:
                        as_str = list(str(x) for x in y)
                        interior = ' '.join(as_str)
                        current.append('(' + interior + ')')
                    else:
                        raise Exception('format error in sites2str')
                else:
                    raise Exception('format error in sites2str')
            interior = ' or '.join(current)
            processing.append('(' + interior + ')')
        else:
            raise Exception('format error in sites2str')
    return ' '.join(processing)




import re
from types import StringType, ListType, DictType, IntType

from oldowan.polymorphism import Polymorphism
from str2sites import str2sites

# rCRSlist is the rCRS sequence exploded into a list with biological numbering
from oldowan.mtdna import rCRSlist

from oldowan.mtdna import HVR1_indices
from oldowan.mtdna import HVR2_indices
from oldowan.mtdna import HVR1and2_indices
from oldowan.mtdna import HVR1to2_indices
from oldowan.mtdna import coding_indices
from oldowan.mtdna import all_indices

REGIONS = { 'HVR1'    : HVR1_indices, 
            'HVR2'    : HVR2_indices,
            'HVR1AND2': HVR1and2_indices,
            'HVR1TO2' : HVR1to2_indices,
            'CODING'  : coding_indices,
            'ALL'     : all_indices }

VALID_REGIONS = REGIONS.keys()

def flatten(item):
    if type(item) == ListType:
        if len(item) > 1:
            return flatten(item[0]) + flatten(item[1:])
        elif len(item) == 1:
            return flatten(item[0])
        else:
            return []
    else:
        return [item]


def sites2seq(sites, region='HVR1', add16k=False):
    """
    Translate sites to sequence.

    The 'sites' variable can be provided as: 
    1. a string of positions and their values:
        '16129A 16189C 16223T'
    2. a list of strings:
        ['16129A', '16189C', '16223T']
    2. a list of Polymorphisms:
        [Polymorphism(16129,0,'A'), Polymorphism(16189,0,'C'), ... ]

    For the revised Cambridge Reference Sequence, either submit the empty 
    string ('') or 'rCRS' as the sites variable.

    The options for the 'region' argument are:
    1. 'HVR1'
    2. 'HVR2'
    3. 'HVR1and2'
    4. 'HVR1to2'
    5. 'coding'
    6. 'all'
    7. an array of sites. These site numbers are assumed to start at '1'
       following the usual mtDNA site numbering convention. (Not the starts
       at '0' computer science convention.)

    I am following the EMPOP Mitochondrial DNA Control Region Database 
    definition of HVR1 (HVS-I) and HVR2 (HVS-II).

    HVR1 consists of bases [16024-16365] of the revised Cambridge Reference
    Sequence (rCRS), which is available in GenBank under accession number
    AC_000021.2 (gi: 115315570)

    HVR2 consists of bases [73-340] of the rCRS.

    HVR1and2 is HVR1 and HVR2 concatenated.

    HVR1to2 runs from the start of HVR1 around to the end of HVR2.


    Calling sites2seq with an empty string or array will return 'region' from
    the rCRS. The default 'region' is HVR1.

    """
    region_type = 'invalid'
    if type(region) == ListType:
        region_type = 'list'
    elif callable(getattr(region, 'upper')) and region.upper() in VALID_REGIONS:
        region_type = 'string'
        region = region.upper()
    if region_type == 'invalid':
        raise Exception('"region" argument "%s" is invalid' % region)

    # the rCRSlist sequence has had an opening spacer added 
    # so that indexing starts at 1
    working = list(rCRSlist) 

    if (type(sites) == StringType):
        if sites.upper() == 'RCRS':
            sites = ''
        sites = str2sites(sites,add16k=add16k)
    elif (type(sites) == ListType):
        sites = flatten(list(str2sites(x,add16k=add16k) for x in sites))
    else:
        raise Exception('"sites" argument "%s" is not acceptable' % type(sites))

    # need to deal with insertions separately so that the index numbering
    # can be maintained
    insertions = []
    
    for poly in sites:
        if poly.position < 1 or poly.position > 16569:
            msg = "'%s' is not in the human mtDNA sequence range" % poly
            raise Exception, msg
        elif poly.insert > 0:
            insertions.append(poly)
        else:
            working[poly.position] = poly.value

    requested_positions = None
    if region_type == 'string':
        requested_positions = REGIONS[region]
    else:
        requested_positions = region

    # as insertions are added to the working sequence, we have to modify the
    # requested_positions to account for changes in the length of the sequence
    if len(insertions) > 0:
        # working from highest position to lowest means that we don't have to
        # keep updating the position of the insertion
        insertions.sort(reverse=True)
        for current in insertions:
            increment = lambda x: x + 1 if x > current.position else x
            requested_positions = list(increment(x) for x in requested_positions)
            # the list completion properly increments all positions higher than
            # the insertion position, but ends up skipping the point of
            # insertion altogether, so add that back in to the list
            insert_index = requested_positions.index(current.position)+1
            requested_positions.insert(insert_index, current.position+1)
            working.insert(current.position+1, current.value)
        
    result = list(working[x] for x in requested_positions)
    return ''.join(result)




from oldowan.polymorphism import Polymorphism
from mitomap_data import substitutions

# COMMON ABBREVIATIONS USED HERE
#
# mb: mismatch block, a list of lists
#     each sublist holds the Polymorphisms for an alternate alignment
#

#
# UTILITY FUNCTIONS
#

def cmp_alternate_alignments(a,b):
    """Compare function for lists of Polymorphisms.

    Comparison is based upon Polymorphism positions.
    """
    a.sort()
    b.sort()
    for i in range(len(a)-1,-1,-1):
        if a[i].position != b[i].position:
            return cmp(a[i].position, b[i].position)
        if a[i].insert != b[i].insert:
            return cmp(a[i].insert, b[i].insert)
    return 0

#
# REDUCTION FUNCTIONS
#

def prefer_fewer(mb):
    """Prefer fewer polymorphisms."""
    mb.sort(cmp=lambda x,y: cmp(len(x),len(y)))
    mb = list(x for x in mb if len(x) == len(mb[0]))
    return mb


def prefer_known_substitutions(mb):
    """Prefer substitutions previously seen to new substitutions.

    Previously seen is based upon the listing from www.mitomap.org.
    Each alternate alignment receives a score calculated based on
    how many of the substitutions in that alignment are in the
    mitomap variants list: 1 point for transversions and 2 points 
    for transitions.
    """
    if len(mb) > 1:
        scores = [0] * len(mb)
        # pos: position
        # aa: alternate alignment
        for pos, aa in enumerate(mb):
            for variant in aa:
                if variant.is_substitution():
                    if str(variant) in substitutions:
                        # larger scores are "better"
                        if variant.is_transition():
                            scores[pos] += 2
                        else:
                            scores[pos] += 1
        if max(scores) > 0:
            # lsi: indices of lower scoring alternate alignments
            lsi = list(x for x in range(len(mb)) if scores[x] < max(scores))
            # remove low scoring alignments from the mismatch block
            # in reverse order so as to not mess up the preceding indices
            lsi.sort(reverse=True)
            for i in lsi:
                mb.pop(i)

    return mb


def prefer_multi_inserts(mb):
    """Prefer several insertions at one site over separate insertions.

    When there is a choice between multiple inserts at one position
    or multiple inserts spread over several positions, choose
    multiple inserts at the same position.
    """
    # position_counts: number of insertion sites required in 
    #                  each alternate alignment
    position_counts = {}
    # pos: position
    # aa: alternate alignment
    for pos,aa in enumerate(mb):
        inserts = list(x for x in aa if x.insert > 0)
        positions = len(set(list(x.position for x in inserts)))
        position_counts[pos] = positions

    # if there are no insertions, this rule does not
    # apply, and there will be no position counts > 0
    if len(position_counts.values()) > 0:
        # minpos: minimum number of insertion sites
        minpos = min(position_counts.values())
    
        # only continue if there is any variation in number of
        # insertion positions required.
        if minpos != max(position_counts.values()):
            new_mb = []
            for i in range(len(mb)):
                if position_counts[i] == minpos:
                    new_mb.append(mb[i])
            mb = new_mb

    return mb


def prefer_indels_at_end(mb):
    """Prefer alignments with insertions and deletions moved 3'.

    For example, if an insertion could be placed in more than one position,
    choose the most 3' (3 prime) position. Here, the reference sequence has 
    two C's in a row, but the submitted sequence has three, so there are 3 
    possible alignments:

        Reference  : GA-CCTT
        Alignment 1: GACCCTT

        Reference  : GAC-CTT
        Alignment 2: GACCCTT

        Reference  : GACC-TT
        Alignment 3: GACCCTT

    We should choose Alignment 3 as this alternative has the 
    insertion in the most 3' possible position.
    """
    # don't waste time if there are no alternatives
    if len(mb) <= 1:
        return mb

    # indel_set_flag: set to True if the first alternate alignment
    #                 includes at least one indel. Falsified if later
    #                 alignments have a different number of indels than
    #                 the first. This rule will then not be applied.
    indel_set_flag = False

    # indel_set_len: the number of indels in the first alternative
    #                alignment. The others must have the same number.
    indel_set_len = None

    # indel_sets: the lists of indels in each alternate alignment is
    #             collected here
    indel_sets = []

    # aa: alternate alignment
    for aa in mb:

        indels = list(x for x in aa if x.insert > 0 or x.value == '-')

        # initialize the size of the insertion set
        if indel_set_len is None and len(indels) > 0:
            indel_set_flag = True
            indel_set_len = len(indels)

        # extract out sets of insertions
        if indel_set_flag and len(indels) == indel_set_len:
            indels.sort()
            indel_sets.append(list(x for x in indels))
        else:
            indel_set_flag = False

    # only apply the rule if we have found alternatives with indels
    if indel_set_len > 0 and indel_set_flag:
        indel_sets.sort(cmp=cmp_alternate_alignments)

        # full_sets: the whole alternative alignment, including substitutions
        full_sets = []
        for aa in mb:
            if indel_sets[-1][-1] in aa:
                full_sets.append(aa)

        full_sets.sort(cmp=cmp_alternate_alignments)

        return [full_sets[-1]]
        #return list(full_sets[-1] for i in range(len(mb)))
        
    return mb


def prefer_insertions_at_309_and_315(mb):
    """Prefer alternatives that include 309.1C or 315.1C over others.

    There are two multi-C runs at the beginning of the 300's and by
    convention, any insert in one of those runs is pushed to the end.
    Mostly, the other rules will pick these, but in some circumstances
    with other substitutions or deletions in this area, these won't get
    picked - although we want them to. Thus, this special case preference.
    """
    special_cases = ['309.1C', '315.1C']
    # mb: mismatch block 
    if len(mb) > 1:
        scores = [0] * len(mb)
        # pos: position
        # aa: alternate alignment
        for pos, aa in enumerate(mb):
            for variant in aa:
                if str(variant) in special_cases:
                    scores[pos] += 1
        if max(scores) > 0:
            # lsi: indices of lower scoring alternate alignments
            lsi = list(x for x in range(len(mb)) if scores[x] < max(scores))
            # remove low scoring alignments from the mismatch block
            # in reverse order so as to not mess up the preceding indices
            lsi.sort(reverse=True)
            for i in lsi:
                mb.pop(i)

    return mb


def prefer_315_insertion_over_double_310_insertion(mb):
    """Solely to (partially) fulfill Wilson et al. 2002 example 19.

    Wilson et al's solution to their example 19 fails to adhere to
    their rules. They suggest that the proper alignment should be:

        REF: CCT-CCCCC-GCT
        OBS: CCTTCCCCCCGCT

    However, their rule 3 is that insertions and deletions should
    be combined where the same number of differences to the reference
    sequence is maintained.

    Thus, following their rules, the alignment should be:

        REF: CCT--CCCCCGCT
        OBS: CCTTCCCCCCGCT

    However, Wilson et al 2002 have an unwritten rule 0:

        Rule 0: Prefer known variants.

    Thus, because the insertion placement rules will almost always
    put an insertion in the second run of C's above at 315.1C, as 
    in their preferred alignment, we will do the same here.

    However, any universal application of a rule that will do this
    breaks a lot of other desired behavior, thus, this special pleading.
    """
    before = ['310.1T', '310.2C']
    if len(mb) == 1 and len(mb[0]) == 2:
        var1, var2 = mb[0][0], mb[0][1]
        if str(var2) in before and str(var2) in before:
            var1 = Polymorphism(310,1,'T','-')
            var2 = Polymorphism(315,1,'C','-')
            return [[var1,var2]]
    return mb


def prefer_309del_315ins_over_309T_310C(mb):
    """Fulfill Wilson et al. 2002 example 18.

    Pairwise alignment parameters must favor substitutions over
    indels for most outcomes to be what is expected. However, there
    are situations, like this, where that mismatch between what must
    be done algorithmically conflicts with the selection rules (which
    state that indels should be preferred over substitutions). Thus,
    we have this special case.

    Without this special case, the solution to Wilson et al's example
    18 is 309T, 310C:

        REF: AAACCCCCCCTCCCCCGC
        OBS: AAACCCCCCTCCCCCCGC

    But, following the rule, that indels are preferred over substitutions,
    Wilson et al offer 309d, 315.1C:

        REF: AAACCCCCCCTCCCCC-GC
        OBS: AAACCCCCC-TCCCCCCGC

    However, their rule 3 is that insertions and deletions should
    be preferred where the same number of differences to the reference
    sequence is maintained.

    Therefore, this rule replaces 309T, 310C with 309d, 3151.C

    Any universal application of a rule that will do this
    breaks a lot of other desired behavior, thus, this special pleading.
    """
    before = ['309T', '310C']
    if len(mb) == 1 and len(mb[0]) == 2:
        var1, var2 = mb[0][0], mb[0][1]
        if str(var2) in before and str(var2) in before:
            var1 = Polymorphism(309,0,'-','C')
            var2 = Polymorphism(315,1,'C','-')
            return [[var1,var2]]
    return mb


def prefer_95ins_97del_over_96T_97C(mb):
    """Fulfill Wilson et al. 2002 example 22.

    Pairwise alignment parameters must favor substitutions over
    indels for most outcomes to be what is expected. However, there
    are situations, like this, where that mismatch between what must
    be done algorithmically conflicts with the selection rules (which
    state that indels should be preferred over substitutions). Thus,
    we have this special case.

    Without this special case, the solution to Wilson et al's example
    18 is 96T, 97C:

        REF: AGACGCTGGAGCC
        OBS: AGATCCTGGAGCC

    But, following the rule, that indels are preferred over substitutions,
    Wilson et al offer 95.1T, 97d:

        REF: AGA-CGCTGGAGCC
        OBS: AGATC-CTGGAGCC

    However, their rule 3 is that insertions and deletions should
    be preferred where the same number of differences to the reference
    sequence is maintained.

    Therefore, this rule replaces 96T, 97C with 95.1T,97d 

    Any universal application of a rule that will do this
    breaks a lot of other desired behavior, thus, this special pleading.
    """
    if len(mb) == 1:
        mb_str = list(str(x) for x in mb[0])
        if '96T' in mb_str and '97C' in mb_str:
            var1_ind = mb_str.index('96T')
            var2_ind = mb_str.index('97C')
            var1 = Polymorphism(95,1,'T','-')
            var2 = Polymorphism(97,0,'-','G')
            mb[0][var1_ind] = var1
            mb[0][var2_ind] = var2
    return mb


def prefer_16183C_16188T_16189C_over_16183d_16193insC(mb):
    """16183C and 16189C are common substitutions that are lost sometimes.

    This special case breaks the rules (more changes than necessary),
    but keeps common substitutions 16183C and 16189C, which is observed
    in 655 out of 5049 haplotypes in the 2007 Genographic haplotype listing.

    In this particular example, there is the 16188T substitution, which 
    leads the algorithm to choose 16183d, 16193.1C.
    """
    for aa in mb:
        aa_str = list(str(x) for x in aa)
        a = '16183d'
        b = '16193.1C'
        if a in aa_str and b in aa_str:
            var1_ind = aa_str.index(a)
            var2_ind = aa_str.index(b)
            indices = [var1_ind, var2_ind]
            indices.sort(reverse=True)
            for i in indices:
                aa.pop(i)
            aa.append(Polymorphism(16183,0,'C','A'))
            aa.append(Polymorphism(16188,0,'T','C'))
            aa.append(Polymorphism(16189,0,'C','T'))
            aa.sort() 
    return mb



def prefer_indels_over_ts_over_tv(mb):
    """Prefer indels over transitions over transversions.

    Following reccomendation 2 of Wilson et al. 2002:
       
       If there is more than one way to maintain the same number of 
       differences with respect to the reference sequence, differences
       should be prioritized in the following manner:
       (a) insertions/deletions (indels)
       (b) transitions
       (c) transversions

    """
    if len(mb) > 1:
        scores = [0] * len(mb)
        # pos: position
        # aa: alternate alignment
        for pos, aa in enumerate(mb):
            for variant in aa:
                if variant.is_insertion() or variant.is_deletion():
                    scores[pos] += 2
                elif variant.is_transition():
                    scores[pos] += 1
        if max(scores) > 0:
            # lsi: indices of lower scoring alternate alignments
            lsi = list(x for x in range(len(mb)) if scores[x] < max(scores))
            # remove low scoring alignments from the mismatch block
            # in reverse order so as to not mess up the preceding indices
            lsi.sort(reverse=True)
            for i in lsi:
                mb.pop(i)

    return mb


def prefer_end_of_sorted_list(mb):
    """When all eles fails, prefer the end of the sorted list."""
    for aa in mb:
        aa.sort()
    mb.sort()
    return [mb[-1]]
